"""
Functions which convert between various types of color values.

"""

# SPDX-License-Identifier: BSD-3-Clause

from ._definitions import CSS3, _get_hex_to_name_map, _get_name_to_hex_map
from ._normalization import (
    _percent_to_integer,
    normalize_hex,
    normalize_integer_triplet,
    normalize_percent_triplet,
)
from ._types import IntegerRGB, IntTuple, PercentRGB, PercentTuple

# Conversions from color names to other formats.
# --------------------------------------------------------------------------------


def name_to_hex(name: str, spec: str = CSS3) -> str:
    """
    Convert a color name to a normalized hexadecimal color value.

    The color name will be normalized to lower-case before being looked up.

    Examples:

    .. doctest::

        >>> name_to_hex("white")
        '#ffffff'
        >>> name_to_hex("navy")
        '#000080'
        >>> name_to_hex("goldenrod")
        '#daa520'
        >>> name_to_hex("goldenrod", spec=HTML4)
        Traceback (most recent call last):
            ...
        ValueError: "goldenrod" is not defined as a named color in html4.

    :param name: The color name to convert.
    :param spec: The specification from which to draw the list of color names. Default
       is :data:`CSS3`.
    :raises ValueError: when the given name has no definition in the given spec.

    """
    color_map = _get_name_to_hex_map(spec)
    if hex_value := color_map.get(name.lower()):
        return hex_value
    raise ValueError(f'"{name}" is not defined as a named color in {spec}')


def name_to_rgb(name: str, spec: str = CSS3) -> IntegerRGB:
    """
    Convert a color name to a 3-:class:`tuple` of :class:`int` suitable for use in
    an ``rgb()`` triplet specifying that color.

    The color name will be normalized to lower-case before being looked up.

    Examples:

    .. doctest::

        >>> name_to_rgb("white")
        IntegerRGB(red=255, green=255, blue=255)
        >>> name_to_rgb("navy")
        IntegerRGB(red=0, green=0, blue=128)
        >>> name_to_rgb("goldenrod")
        IntegerRGB(red=218, green=165, blue=32)

    :param name: The color name to convert.
    :param spec: The specification from which to draw the list of color names. Default
       is :data:`CSS3.`
    :raises ValueError: when the given name has no definition in the given spec.

    """
    return hex_to_rgb(name_to_hex(name, spec=spec))


def name_to_rgb_percent(name: str, spec: str = CSS3) -> PercentRGB:
    """
    Convert a color name to a 3-:class:`tuple` of percentages suitable for use in an
    ``rgb()`` triplet specifying that color.

    The color name will be normalized to lower-case before being looked up.

    Examples:

    .. doctest::

        >>> name_to_rgb_percent("white")
        PercentRGB(red='100%', green='100%', blue='100%')
        >>> name_to_rgb_percent("navy")
        PercentRGB(red='0%', green='0%', blue='50%')
        >>> name_to_rgb_percent("goldenrod")
        PercentRGB(red='85.49%', green='64.71%', blue='12.5%')

    :param name: The color name to convert.
    :param spec: The specification from which to draw the list of color names. Default
       is :data:`CSS3`.
    :raises ValueError: when the given name has no definition in the given spec.

    """
    return rgb_to_rgb_percent(name_to_rgb(name, spec=spec))


# Conversions from hexadecimal color values to other formats.
# --------------------------------------------------------------------------------


def hex_to_name(hex_value: str, spec: str = CSS3) -> str:
    """
    Convert a hexadecimal color value to its corresponding normalized color name, if
    any such name exists.

    The hexadecimal value will be normalized before being looked up.

    .. note:: **Spelling variants**

       Some values representing named gray colors can map to either of two names in
       CSS3, because it supports both ``"gray"`` and ``"grey"`` spelling variants for
       those colors. This function will always return the variant spelled ``"gray"``
       (such as ``"lightgray"`` instead of ``"lightgrey"``). See :ref:`the documentation
       on name conventions <color-name-conventions>` for details.

    Examples:

    .. doctest::

        >>> hex_to_name("#ffffff")
        'white'
        >>> hex_to_name("#fff")
        'white'
        >>> hex_to_name("#000080")
        'navy'
        >>> hex_to_name("#daa520")
        'goldenrod'
        >>> hex_to_name("#daa520", spec=HTML4)
        Traceback (most recent call last):
            ...
        ValueError: "#daa520" has no defined color name in html4.

    :param hex_value: The hexadecimal color value to convert.
    :param spec: The specification from which to draw the list of color names. Default
       is :data:`CSS3`.
    :raises ValueError: when the given color has no name in the given spec, or when the
       supplied hex value is invalid.

    """
    color_map = _get_hex_to_name_map(spec)
    if name := color_map.get(normalize_hex(hex_value)):
        return name
    raise ValueError(f'"{hex_value}" has no defined color name in {spec}.')


def hex_to_rgb(hex_value: str) -> IntegerRGB:
    """
    Convert a hexadecimal color value to a 3-:class:`tuple` of :class:`int` suitable
    for use in an ``rgb()`` triplet specifying that color.

    The hexadecimal value will be normalized before being converted.

    Examples:

    .. doctest::

        >>> hex_to_rgb("#fff")
        IntegerRGB(red=255, green=255, blue=255)
        >>> hex_to_rgb("#000080")
        IntegerRGB(red=0, green=0, blue=128)

    :param hex_value: The hexadecimal color value to convert.
    :raises ValueError: when the supplied hex value is invalid.

    """
    int_value = int(normalize_hex(hex_value)[1:], 16)
    return IntegerRGB(int_value >> 16, int_value >> 8 & 0xFF, int_value & 0xFF)


def hex_to_rgb_percent(hex_value: str) -> PercentRGB:
    """
    Convert a hexadecimal color value to a 3-:class:`tuple` of percentages suitable
    for use in an ``rgb()`` triplet representing that color.

    The hexadecimal value will be normalized before being converted.

    Examples:

    .. doctest::

        >>> hex_to_rgb_percent("#ffffff")
        PercentRGB(red='100%', green='100%', blue='100%')
        >>> hex_to_rgb_percent("#000080")
        PercentRGB(red='0%', green='0%', blue='50%')

    :param hex_value: The hexadecimal color value to convert.
    :raises ValueError: when the supplied hex value is invalid.

    """
    return rgb_to_rgb_percent(hex_to_rgb(hex_value))


# Conversions from  integer rgb() triplets to other formats.
# --------------------------------------------------------------------------------


def rgb_to_name(rgb_triplet: IntTuple, spec: str = CSS3) -> str:
    """
    Convert a 3-:class:`tuple` of :class:`int`, suitable for use in an ``rgb()``
    color triplet, to its corresponding normalized color name, if any such name exists.

    To determine the name, the triplet will be converted to a normalized hexadecimal
    value.

    .. note:: **Spelling variants**

       Some values representing named gray colors can map to either of two names in
       CSS3, because it supports both ``"gray"`` and ``"grey"`` spelling variants for
       those colors. This function will always return the variant spelled ``"gray"``
       (such as ``"lightgray"`` instead of ``"lightgrey"``). See :ref:`the documentation
       on name conventions <color-name-conventions>` for details.

    Examples:

    .. doctest::

        >>> rgb_to_name((255, 255, 255))
        'white'
        >>> rgb_to_name((0, 0, 128))
        'navy'

    :param rgb_triplet: The ``rgb()`` triplet.
    :param spec: The specification from which to draw the list of color names. Default
       is :data:`CSS3`.
    :raises ValueError: when the given color has no name in the given spec.

    """
    return hex_to_name(rgb_to_hex(normalize_integer_triplet(rgb_triplet)), spec=spec)


def rgb_to_hex(rgb_triplet: IntTuple) -> str:
    """
    Convert a 3-:class:`tuple` of :class:`int`, suitable for use in an ``rgb()``
    color triplet, to a normalized hexadecimal value for that color.

    Examples:

    .. doctest::

        >>> rgb_to_hex((255, 255, 255))
        '#ffffff'
        >>> rgb_to_hex((0, 0, 128))
        '#000080'

    :param rgb_triplet: The ``rgb()`` triplet.

    """
    red, green, blue = normalize_integer_triplet(rgb_triplet)
    return f"#{red:02x}{green:02x}{blue:02x}"


def rgb_to_rgb_percent(rgb_triplet: IntTuple) -> PercentRGB:
    """
    Convert a 3-:class:`tuple` of :class:`int`, suitable for use in an ``rgb()``
    color triplet, to a 3-:class:`tuple` of percentages suitable for use in representing
    that color.

    .. note:: **Floating-point precision**

       This function makes some trade-offs in terms of the accuracy of the final
       representation. For some common integer values, special-case logic is used to
       ensure a precise result (e.g., integer 128 will always convert to ``"50%"``,
       integer 32 will always convert to ``"12.5%"``), but for all other values a
       standard Python :class:`float` is used and rounded to two decimal places, which
       may result in a loss of precision for some values due to the inherent imprecision
       of `IEEE floating-point numbers <https://en.wikipedia.org/wiki/IEEE_754>`_.

    Examples:

    .. doctest::

        >>> rgb_to_rgb_percent((255, 255, 255))
        PercentRGB(red='100%', green='100%', blue='100%')
        >>> rgb_to_rgb_percent((0, 0, 128))
        PercentRGB(red='0%', green='0%', blue='50%')
        >>> rgb_to_rgb_percent((218, 165, 32))
        PercentRGB(red='85.49%', green='64.71%', blue='12.5%')

    :param rgb_triplet: The ``rgb()`` triplet.

    """
    # In order to maintain precision for common values,
    # special-case them.
    specials = {
        255: "100%",
        128: "50%",
        64: "25%",
        32: "12.5%",
        16: "6.25%",
        0: "0%",
    }
    return PercentRGB._make(
        specials.get(d, f"{d / 255.0 * 100:.02f}%")
        for d in normalize_integer_triplet(rgb_triplet)
    )


# Conversions from percentage rgb() triplets to other formats.
# --------------------------------------------------------------------------------


def rgb_percent_to_name(rgb_percent_triplet: PercentTuple, spec: str = CSS3) -> str:
    """
    Convert a 3-:class:`tuple` of percentages, suitable for use in an ``rgb()``
    color triplet, to its corresponding normalized color name, if any such name exists.

    To determine the name, the triplet will be converted to a normalized hexadecimal
    value.

    .. note:: **Spelling variants**

       Some values representing named gray colors can map to either of two names in
       CSS3, because it supports both ``"gray"`` and ``"grey"`` spelling variants for
       those colors. This function will always return the variant spelled ``"gray"``
       (such as ``"lightgray"`` instead of ``"lightgrey"``). See :ref:`the documentation
       on name conventions <color-name-conventions>` for details.

    Examples:

    .. doctest::

        >>> rgb_percent_to_name(("100%", "100%", "100%"))
        'white'
        >>> rgb_percent_to_name(("0%", "0%", "50%"))
        'navy'
        >>> rgb_percent_to_name(("85.49%", "64.71%", "12.5%"))
        'goldenrod'

    :param rgb_percent_triplet: The ``rgb()`` triplet.
    :param spec: The specification from which to draw the list of color names. Default
        is :data:`CSS3`.
    :raises ValueError: when the given color has no name in the given spec.

    """
    return rgb_to_name(
        rgb_percent_to_rgb(normalize_percent_triplet(rgb_percent_triplet)),
        spec=spec,
    )


def rgb_percent_to_hex(rgb_percent_triplet: PercentTuple) -> str:
    """
    Convert a 3-:class:`tuple` of percentages, suitable for use in an ``rgb()``
    color triplet, to a normalized hexadecimal color value for that color.

    Examples:

    .. doctest::

        >>> rgb_percent_to_hex(("100%", "100%", "0%"))
        '#ffff00'
        >>> rgb_percent_to_hex(("0%", "0%", "50%"))
        '#000080'
        >>> rgb_percent_to_hex(("85.49%", "64.71%", "12.5%"))
        '#daa520'

    :param rgb_percent_triplet: The ``rgb()`` triplet.

    """
    return rgb_to_hex(
        rgb_percent_to_rgb(normalize_percent_triplet(rgb_percent_triplet))
    )


def rgb_percent_to_rgb(rgb_percent_triplet: PercentTuple) -> IntegerRGB:
    """
    Convert a 3-:class:`tuple` of percentages, suitable for use in an ``rgb()``
    color triplet, to a 3-:class:`tuple` of :class:`int` suitable for use in
    representing that color.

    Some precision may be lost in this conversion. See the note regarding precision for
    :func:`~webcolors.rgb_to_rgb_percent` for details.

    Examples:

    .. doctest::

        >>> rgb_percent_to_rgb(("100%", "100%", "100%"))
        IntegerRGB(red=255, green=255, blue=255)
        >>> rgb_percent_to_rgb(("0%", "0%", "50%"))
        IntegerRGB(red=0, green=0, blue=128)
        >>> rgb_percent_to_rgb(("85.49%", "64.71%", "12.5%"))
        IntegerRGB(red=218, green=165, blue=32)

    :param rgb_percent_triplet: The ``rgb()`` triplet.

    """
    return IntegerRGB._make(
        map(
            _percent_to_integer,  # pylint: disable=protected-access
            normalize_percent_triplet(rgb_percent_triplet),
        )
    )



"""
HTML5 color algorithms.

Note that these functions are written in a way that may seem strange to developers
familiar with Python, because they do not use the most efficient or idiomatic way of
accomplishing their tasks. This is because, for compliance, these functions are written
as literal translations into Python of the algorithms in HTML5:

https://html.spec.whatwg.org/multipage/common-microsyntaxes.html#colours

For ease of understanding, the relevant steps of the algorithm from the standard are
included as comments interspersed in the implementation.

"""

# SPDX-License-Identifier: BSD-3-Clause

import string

from ._definitions import _CSS3_NAMES_TO_HEX
from ._types import HTML5SimpleColor, IntTuple


def html5_parse_simple_color(value: str) -> HTML5SimpleColor:
    """
    Apply the HTML5 simple color parsing algorithm.

    Examples:

    .. doctest::

        >>> html5_parse_simple_color("#ffffff")
        HTML5SimpleColor(red=255, green=255, blue=255)
        >>> html5_parse_simple_color("#fff")
        Traceback (most recent call last):
            ...
        ValueError: An HTML5 simple color must be a string seven characters long.

    :param value: The color to parse.
    :type value: :class:`str`, which must consist of exactly the character ``"#"``
        followed by six hexadecimal digits.
    :raises ValueError: when the given value is not a Unicode string of length 7,
       consisting of exactly the character ``#`` followed by six hexadecimal digits.

    """
    # 1. Let input be the string being parsed.
    #
    # 2. If input is not exactly seven characters long, then return an error.
    if not isinstance(value, str) or len(value) != 7:
        raise ValueError(
            "An HTML5 simple color must be a Unicode string seven characters long."
        )

    # 3. If the first character in input is not a U+0023 NUMBER SIGN character (#), then
    #    return an error.
    if not value.startswith("#"):
        raise ValueError(
            "An HTML5 simple color must begin with the character '#' (U+0023)."
        )

    # 4. If the last six characters of input are not all ASCII hex digits, then return
    #    an error.
    if not all(c in string.hexdigits for c in value[1:]):
        raise ValueError(
            "An HTML5 simple color must contain exactly six ASCII hex digits."
        )

    # 5. Let result be a simple color.
    #
    # 6. Interpret the second and third characters as a hexadecimal number and let the
    #    result be the red component of result.
    #
    # 7. Interpret the fourth and fifth characters as a hexadecimal number and let the
    #    result be the green component of result.
    #
    # 8. Interpret the sixth and seventh characters as a hexadecimal number and let the
    #    result be the blue component of result.
    #
    # 9. Return result.
    return HTML5SimpleColor(
        int(value[1:3], 16), int(value[3:5], 16), int(value[5:7], 16)
    )


def html5_serialize_simple_color(simple_color: IntTuple) -> str:
    """
    Apply the HTML5 simple color serialization algorithm.

    Examples:

    .. doctest::

        >>> html5_serialize_simple_color((0, 0, 0))
        '#000000'
        >>> html5_serialize_simple_color((255, 255, 255))
        '#ffffff'

    :param simple_color: The color to serialize.

    """
    red, green, blue = simple_color

    # 1. Let result be a string consisting of a single "#" (U+0023) character.
    #
    # 2. Convert the red, green, and blue components in turn to two-digit hexadecimal
    #    numbers using lowercase ASCII hex digits, zero-padding if necessary, and append
    #    these numbers to result, in the order red, green, blue.
    #
    # 3. Return result, which will be a valid lowercase simple color.
    return f"#{red:02x}{green:02x}{blue:02x}"


def html5_parse_legacy_color(value: str) -> HTML5SimpleColor:
    """
    Apply the HTML5 legacy color parsing algorithm.

    Note that, since this algorithm is intended to handle many types of
    malformed color values present in real-world Web documents, it is
    *extremely* forgiving of input, but the results of parsing inputs
    with high levels of "junk" (i.e., text other than a color value)
    may be surprising.

    Examples:

    .. doctest::

        >>> html5_parse_legacy_color("black")
        HTML5SimpleColor(red=0, green=0, blue=0)
        >>> html5_parse_legacy_color("chucknorris")
        HTML5SimpleColor(red=192, green=0, blue=0)
        >>> html5_parse_legacy_color("Window")
        HTML5SimpleColor(red=0, green=13, blue=0)

    :param value: The color to parse.

    :raises ValueError: when the given value is not a Unicode string, when it is the
       empty string, or when it is precisely the string ``"transparent"``.

    """
    # 1. Let input be the string being parsed.
    if not isinstance(value, str):
        raise ValueError(
            "HTML5 legacy color parsing requires a Unicode string as input."
        )

    # 2. If input is the empty string, then return an error.
    if value == "":
        raise ValueError("HTML5 legacy color parsing forbids empty string as a value.")

    # 3. Strip leading and trailing ASCII whitespace from input.
    value = value.strip()

    # 4. If input is an ASCII case-insensitive match for the string "transparent", then
    #    return an error.
    if value.lower() == "transparent":
        raise ValueError('HTML5 legacy color parsing forbids "transparent" as a value.')

    # 5. If input is an ASCII case-insensitive match for one of the named colors, then
    #    return the simple color corresponding to that keyword.
    #
    #    Note: CSS2 System Colors are not recognized.
    if keyword_hex := _CSS3_NAMES_TO_HEX.get(value.lower()):
        return html5_parse_simple_color(keyword_hex)

    # 6. If input's code point length is four, and the first character in input is
    #    U+0023 (#), and the last three characters of input are all ASCII hex digits,
    #    then:
    if (
        len(value) == 4
        and value.startswith("#")
        and all(c in string.hexdigits for c in value[1:])
    ):
        # 1. Let result be a simple color.
        #
        # 2. Interpret the second character of input as a hexadecimal digit; let the red
        #    component of result be the resulting number multiplied by 17.
        #
        # 3. Interpret the third character of input as a hexadecimal digit; let the
        #    green component of result be the resulting number multiplied by 17.
        #
        # 4. Interpret the fourth character of input as a hexadecimal digit; let the
        #    blue component of result be the resulting number multiplied by 17.
        result = HTML5SimpleColor(
            int(value[1], 16) * 17, int(value[2], 16) * 17, int(value[3], 16) * 17
        )

        # 5. Return result.
        return result

    # 7. Replace any code points greater than U+FFFF in input (i.e., any characters that
    #    are not in the basic multilingual plane) with the two-character string "00".
    value = "".join("00" if ord(c) > 0xFFFF else c for c in value)

    # 8. If input's code point length is greater than 128, truncate input, leaving only
    #    the first 128 characters.
    if len(value) > 128:
        value = value[:128]

    # 9. If the first character in input is a U+0023 NUMBER SIGN character (#), remove
    #    it.
    if value.startswith("#"):
        value = value[1:]

    # 10. Replace any character in input that is not an ASCII hex digit with the
    # character U+0030 DIGIT ZERO (0).
    value = "".join(c if c in string.hexdigits else "0" for c in value)

    # 11. While input's code point length is zero or not a multiple of three, append a
    #     U+0030 DIGIT ZERO (0) character to input.
    while (len(value) == 0) or (len(value) % 3 != 0):
        value += "0"

    # 12. Split input into three strings of equal code point length, to obtain three
    #     components. Let length be the code point length that all of those components
    #     have (one third the code point length of input).
    length = int(len(value) / 3)
    red = value[:length]
    green = value[length : length * 2]
    blue = value[length * 2 :]

    # 13. If length is greater than 8, then remove the leading length-8 characters in
    #     each component, and let length be 8.
    if length > 8:
        red, green, blue = (red[length - 8 :], green[length - 8 :], blue[length - 8 :])
        length = 8

    # 14. While length is greater than two and the first character in each component is
    #     a U+0030 DIGIT ZERO (0) character, remove that character and reduce length by
    #     one.
    while (length > 2) and (red[0] == "0" and green[0] == "0" and blue[0] == "0"):
        red, green, blue = (red[1:], green[1:], blue[1:])
        length -= 1

    # 15. If length is still greater than two, truncate each component, leaving only the
    #     first two characters in each.
    if length > 2:
        red, green, blue = (red[:2], green[:2], blue[:2])

    # 16. Let result be a simple color.
    #
    # 17. Interpret the first component as a hexadecimal number; let the red component
    #     of result be the resulting number.
    #
    # 18. Interpret the second component as a hexadecimal number; let the green
    #     component of result be the resulting number.
    #
    # 19. Interpret the third component as a hexadecimal number; let the blue component
    #     of result be the resulting number.
    #
    # 20. Return result.
    return HTML5SimpleColor(int(red, 16), int(green, 16), int(blue, 16))



"""
Types and type aliases used to represent colors in various formats.

"""

# SPDX-License-Identifier: BSD-3-Clause

import typing


class IntegerRGB(typing.NamedTuple):
    """
    :class:`~typing.NamedTuple` representing an integer RGB triplet.

    Has three fields, each of type :class:`int` and in the range 0-255 inclusive:

    .. attribute:: red

       The red portion of the color value.

    .. attribute:: green

       The green portion of the color value.

    .. attribute:: blue

       The blue portion of the color value.

    """

    red: int
    green: int
    blue: int


class PercentRGB(typing.NamedTuple):
    """
    :class:`~typing.NamedTuple` representing a percentage RGB triplet.

    Has three fields, each of type :class:`str` and representing a percentage value in
    the range 0%-100% inclusive:

    .. attribute:: red

       The red portion of the color value.

    .. attribute:: green

       The green portion of the color value.

    .. attribute:: blue

       The blue portion of the color value.

    """

    red: str
    green: str
    blue: str


class HTML5SimpleColor(typing.NamedTuple):
    """
    :class:`~typing.NamedTuple` representing an HTML5 simple color.

    Has three fields, each of type :class:`int` and in the range 0-255 inclusive:

    .. attribute:: red

       The red portion of the color value.

    .. attribute:: green

       The green portion of the color value.

    .. attribute:: blue

       The blue portion of the color value.

    """

    red: int
    green: int
    blue: int


# Union type representing the possible types of an integer RGB tuple.
IntTuple = typing.Union[IntegerRGB, HTML5SimpleColor, typing.Tuple[int, int, int]]

# Union type representing the possible types of a percentage RGB tuple.
PercentTuple = typing.Union[PercentRGB, typing.Tuple[str, str, str]]
